<?php

namespace Drupal\orange_dam;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Manages the retrieval and validation of configuration.
 */
class OrangeDamConfigurationManager {

  /**
   * The Orange DAM configuration.
   */
  const ORANGE_DAM_CONFIG = 'orange_dam.settings';

  /**
   * The log channel.
   */
  const LOGGER_CHANNEL = 'orange_dam.content';

  /**
   * Constructor.
   */
  public function __construct(
    protected ConfigFactoryInterface $configFactory,
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
  ) {}

  /**
   * Get an array of allowed content type names.
   *
   * @return string[]
   *   An array of allowed content type names.
   */
  public function getContentTypeNames(): array {
    return array_keys($this->getContentTypes());
  }

  /**
   * Get the content type configuration.
   *
   * @return array
   *   An array of content type configuration.
   */
  public function getContentTypes(): array {
    $contentTypes = $this->configFactory
      ->get(static::ORANGE_DAM_CONFIG)
      ->get('content_types');
    if (empty($contentTypes)) {
      throw new \Exception(
        'Missing "orange_dam.settings content_types" configuration.'
      );
    }
    return $contentTypes;
  }

  /**
   * Get an array of allowed keyword types.
   *
   * @return string[]
   *   An array of allowed content types.
   */
  public function getKeywordTypes(): array {
    $tagTypes = $this->configFactory
      ->get(static::ORANGE_DAM_CONFIG)
      ->get('tag_types');
    if (empty($tagTypes)) {
      throw new \Exception(
        'Missing "orange_dam.settings tag_types" configuration.'
      );
    }
    return $tagTypes;

  }

  /**
   * Get an array of fields to request from the Orange DAM Search API.
   *
   * @return string[]
   *   An array of field names.
   */
  public function getSearchApiFields(): array {
    $searchApiFields = $this->configFactory
      ->get(static::ORANGE_DAM_CONFIG)
      ->get('search_api_fields');
    if (empty($searchApiFields)) {
      throw new \Exception(
        'Missing "orange_dam.settings search_api_fields" configuration.'
      );
    }
    return $searchApiFields;

  }

  /**
   * Get a list of asset formats keyed by their associated content type.
   *
   * @return string[]
   *   A list of asset formats keyed by their associated content type.
   */
  public function getAssetFormats(): array {
    $contentTypes = $this->getContentTypes();
    $contentTypeFormats = [];
    foreach ($contentTypes as $contentType => $contentTypeConfiguration) {
      if (isset($contentTypeConfiguration['asset_format'])) {
        $contentTypeFormats[$contentType] = $contentTypeConfiguration['asset_format'];
      }
    }
    return $contentTypeFormats;
  }

  /**
   * Determine the asset format to use for a given content type.
   *
   * @param string $contentType
   *   The content type.
   *
   * @return string|false
   *   The asset format or FALSE if one was unable to be found.
   *
   * @throws \Exception
   */
  public function getAssetFormat(string $contentType) {
    $assetFormats = $this->getAssetFormats();
    if (isset($assetFormats[$contentType])) {
      return $assetFormats[$contentType];
    }
    // Log a warning if no match was found.
    $this->loggerChannelFactory->get(static::LOGGER_CHANNEL)->warning(
      'Unable to find an asset type mapping for content type :content_type.', [
        ':content_type' => $contentType,
      ]);
    return FALSE;
  }

}
